/*
 * Washer Protector. Controller.
 * main source file
 *
 * (C)2017 St(u)dio of Computer Games
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>
#include <util/delay.h>

#include "wmp.h"
#include "valve.h"
#include "power.h"
#include "leakage.h"
#include "VirtualWire.h"
#include "led.h"
#include "buzzer.h"



#if defined (__AVR_ATmega16__)
#   undef TIMER0_OVF_vect
#   define TIMER0_OVF_vect _VECTOR(18)
#endif // defined (__AVR_ATmega16__)



#define SW_VALVE_CLOSE_PIN      2
#define SW_VALVE_OPEN_PIN       1

#define VW_BAUD                 600

#define ALIVE_PERIOD            60000   // 60s * 1000ucs
#define WASHING_TIME_THR        60 * 5  // 5 minutes
#define WASHING_TIME_ALARM      8       // number of signals about the end of washing

#define WMC_STATE_BATLOW_CLOSE  0
#define WMC_STATE_LEAKAGE_CLOSE 1
#define WMC_STATE_MANUAL_CLOSE  2
#define WMC_STATE_MANUAL_OPEN   3
#define WMC_STATE_REMOTE_CLOSE  4
#define WMC_STATE_REMOTE_OPEN   5
#define WMC_STATE_WTG_CLOSE     6



// debug section
#ifdef DEBUG

//#define DEBUG_VW                1
#define DEBUG_PWR               1
//#define DEBUG_VALVE             1
//#define DEBUG_LK                1

#include <stdio.h>

static void uart_init() {
    UCSR0B |= _BV(RXEN0) | _BV(TXEN0);
	UCSR0C |= _BV(UCSZ00) | _BV(UCSZ01);

#define BAUD 2400
#include <util/setbaud.h>
   UBRR0H = UBRRH_VALUE;
   UBRR0L = UBRRL_VALUE;
#if USE_2X
   UCSR0A |= _BV(U2X0);
#else
   UCSR0A &= ~_BV(U2X0);
#endif
}

static int uart_puthcar(char c, FILE *stream) {
	while ((UCSR0A & _BV(UDRE0)) == 0) {};
	UDR0 = c;
	return 0;
}

static FILE uart_stream = FDEV_SETUP_STREAM(uart_puthcar, NULL, _FDEV_SETUP_RW);


void uart_dump_buffer(uint8_t *data, size_t len) {
    size_t i;

    printf("%d:", len);
    for (i = 0; i < len; i++) {
        printf(" %02x", data[i]);
    }
    printf("\r\n");
}

#endif // DEBUG



//
// scheduler timer
//
static uint16_t cnt_alive_wdg;

typedef struct {
    uint8_t cnt;
    uint8_t sync;
} task_sched_t;

typedef struct {
    uint16_t cnt;
    uint8_t sync;
} task_sched_ext_t;

#define TASK_SCHED_INIT(t)              \
    (t).cnt = 0;                        \
    (t).sync = 0
#define TASK_SCHED_UPDATE(t, period)    \
    if ((t).cnt++ >= period) {          \
        (t).cnt = 0;                    \
        (t).sync = 1;                   \
    }
#define TASK_SCHED_CHECK(t)             \
    (t).sync == 1
#define TASK_SCHED_SYNC(t)              \
    (t).sync = 0


volatile static task_sched_t task_pwr;
volatile static task_sched_t task_lk;
volatile static task_sched_t task_valve;
volatile static task_sched_t task_led;
volatile static task_sched_ext_t task_low;


ISR(TIMER0_OVF_vect) {
    // about every 1.024us

    // ALIVE watchdog
    if (cnt_alive_wdg < ALIVE_PERIOD) cnt_alive_wdg++;

    // POWER shcheduler: about every 8.192us
    TASK_SCHED_UPDATE(task_pwr, 8);

    // LEACKAGE sensor scheduler: about every 249,856us
    TASK_SCHED_UPDATE(task_lk, 244);

    // valve scheduler: about every 100.325us
    TASK_SCHED_UPDATE(task_valve, 98);

    // LED scheduler: about every 10.24us
    TASK_SCHED_UPDATE(task_led, 10);

    // low-priority task: about 1s
    TASK_SCHED_UPDATE(task_low, 977);
}


//
// buzzer sounds
//

void buz_startup() {
    bz_beep(880, 400);
}

void buz_valve_open() {
    bz_beep(110, 300);
    bz_beep(150, 300);
    bz_beep(174, 300);
    bz_beep(230, 300);
}

void buz_valve_close() {
    bz_beep(220, 400);
    bz_beep(117, 800);
}

void buz_valve_error() {
    bz_beep(880, 400);
    _delay_ms(200);
    bz_beep(880, 400);
    _delay_ms(200);
    bz_beep(880, 400);
    _delay_ms(200);
    bz_beep(880, 400);
}

void buz_leak() {
    bz_beep(440, 400);
}

void buz_finish() {
    bz_beep(660, 800);
}



//
// Entry point
//

int main(void) {
    uint8_t wmc_state;
    uint8_t valve_pwr_command;
    uint8_t valve_lk_command;
    uint8_t valve_manual_command;
    uint8_t valve_remote_command;
    uint8_t last_valve_state;
    wmp_msg_t wmp_msg;
    uint8_t msg_len;
    uint16_t washing_time;

#ifdef DEBUG
    uint8_t mcucsr = MCUCSR;
    MCUCSR = 0;
#endif // DEBUG

    // start WDT
    wdt_enable(_BV(WDP3) | _BV(WDP0));
    wdt_reset();

#ifdef DEBUG
    uart_init();
    stdout = &uart_stream;

    printf("WMP Controller, MCUCSR 0x%x\r\n", mcucsr);
#endif // DEBUG

    cli();

    // scheduler 1ucs timer (TIMER2)
    TASK_SCHED_INIT(task_pwr);
    TASK_SCHED_INIT(task_lk);
    TASK_SCHED_INIT(task_valve);
    TASK_SCHED_INIT(task_led);
    TASK_SCHED_INIT(task_low);

    TCCR0B = _BV(CS01) | _BV(CS00);     // OSC / 64
    TIMSK0 = _BV(TOIE0);                // enable Overflow Interrupt

    // FIxMe: SW_VALVE init
    valve_init();
    pwr_init();
    lk_init();
    vw_setup(VW_BAUD);
    led_init();
    bz_init();

    sei();

    // starup beep
    buz_startup();
    _delay_ms(100);

    valve_close();
    vw_rx_start();

    wmc_state = WMC_STATE_WTG_CLOSE;

    valve_pwr_command = VALVE_COMMAND_UNDEFINED;
    valve_lk_command = VALVE_COMMAND_UNDEFINED;
    valve_manual_command = VALVE_COMMAND_UNDEFINED;
    valve_remote_command = VALVE_COMMAND_UNDEFINED;

    last_valve_state = VALVE_STATE_ERROR;

    washing_time = 0;

    while(1) {
        // check POWER status
        if (TASK_SCHED_CHECK(task_pwr)) {
            TASK_SCHED_SYNC(task_pwr);

            pwr_poll();

            uint8_t pwr_state = pwr_get_state();
            if (pwr_state == PWR_STATE_BATLOW) {
                valve_pwr_command = VALVE_COMMAND_CLOSE;
            } else {
                valve_pwr_command = VALVE_COMMAND_UNDEFINED;
            }

            switch (pwr_state) {
            case PWR_STATE_ONLINE:
                led_set(LED_PWR_LINE, LED_ON);
                led_set(LED_PWR_BAT, LED_OFF);
                break;
            case PWR_STATE_ONLINE_CHARGE:
                led_set(LED_PWR_LINE, LED_ON);
                led_set(LED_PWR_BAT, LED_FLASH);
                break;
            case PWR_STATE_BAT:
                led_set(LED_PWR_LINE, LED_OFF);
                led_set(LED_PWR_BAT, LED_ON);
                break;
            case PWR_STATE_BATALARM:
                led_set(LED_PWR_LINE, LED_OFF);
                led_set(LED_PWR_BAT, LED_FLASH);
                break;
            case PWR_STATE_BATLOW:
                break;
            }

#ifdef DEBUG_PWR
            static int i = 0;
            if (i++ >= 128) {
                printf("pwr poll: s %d, c %d l %d, b %d\r\n",
                        pwr_state, pwr_get_charge(), pwr_get_line(), pwr_get_bat());
                i = 0;
            }
#endif // DEBUG
        }

        // check LEAKAGE sensor status
        if (TASK_SCHED_CHECK(task_lk)) {
            TASK_SCHED_SYNC(task_lk);

            lk_poll();

            switch (lk_get_state()) {
            case LK_STATUS_OK:
                valve_lk_command = VALVE_COMMAND_UNDEFINED;
                break;
            case LK_STATUS_ERROR:
                valve_lk_command = VALVE_COMMAND_CLOSE;
                break;
            case LK_STATUS_LEAK:
            default:
                valve_lk_command = VALVE_COMMAND_CLOSE;
                break;
            }
#ifdef DEBUG_LK
            static uint8_t lk_i = 0;
            static uint16_t lk_min = 0xffff;
            static uint16_t lk_max = 0;
            uint16_t lk_val = lk_get_val();

            if (lk_val > lk_max) lk_max = lk_val;
            if (lk_val < lk_min && lk_val > LK_SENSOR_THRESHOLD) lk_min = lk_val;

            if (lk_i++ >= 4) {
                lk_i = 0;
                printf("lk: s %d, v %d (%d - %d)\r\n", lk_get_state(), lk_val, lk_min, lk_max);
            }
#endif // DEBUG
        }

        // read message from RF RX module
        msg_len = sizeof(wmp_msg_t);
        if ((vw_get_message((uint8_t *)&wmp_msg, &msg_len)) && msg_len == sizeof(wmp_msg_t)) {
#ifdef DEBUG_VW
            printf("received message ");
            uart_dump_buffer((uint8_t *)&wmp_msg, msg_len);
#endif // DEBUG_VW

            if (wmp_msg.dst != WMP_ADDR_CONTROLLER || wmp_msg.src != WMP_ADDR_MONITOR) {
#ifdef DEBUG_VW
                printf("bad message: dst %lx, src %lx\r\n", wmp_msg.dst, wmp_msg.src);
#endif // DEBUG_VW
                continue;
            }

            if (!(wmp_msg.status & WMP_MSG_STATUS_ALIVE)) {
#ifdef DEBUG_VW
                printf("not ALIVE message, drop\r\n");
#endif // DEBUG_VW
                continue;
            }

            cnt_alive_wdg = 0;
            led_set(LED_RX, LED_BLINK);

            // check VALVE status
            if (wmp_msg.status & WMP_MSG_STATUS_VALVE) {
#ifdef DEBUG_VW
                printf("open VALVE\r\n");
#endif // DEBUG_VW
                valve_remote_command = VALVE_COMMAND_OPEN;
            } else {
#ifdef DEBUG_VW
                printf("close VALVE\r\n");
#endif // DEBUG_VW
                valve_remote_command = VALVE_COMMAND_CLOSE;
            }
        }

        // reset VALVE remote command if ALIVE watchdog expired
        if (cnt_alive_wdg >= ALIVE_PERIOD) {
            valve_remote_command = VALVE_COMMAND_UNDEFINED;
        }

        // VALVE manual control
        if (!(PINB & _BV(SW_VALVE_CLOSE_PIN)) && (PINB & _BV(SW_VALVE_OPEN_PIN))) {
            valve_manual_command = VALVE_COMMAND_CLOSE;
        } else if ((PINB & _BV(SW_VALVE_CLOSE_PIN)) && !(PINB & _BV(SW_VALVE_OPEN_PIN))) {
            valve_manual_command = VALVE_COMMAND_OPEN;
        } else {
            valve_manual_command = VALVE_COMMAND_UNDEFINED;
        }

        // update VALVE MANUAL LED
        if (wmc_state == WMC_STATE_BATLOW_CLOSE) {
            // switch off VALVE MANUAL LEDs when battery low
            led_set(LED_VALVE_MANUAL, LED_OFF);
        } else if (valve_manual_command == VALVE_COMMAND_UNDEFINED) {
            led_set(LED_VALVE_MANUAL, LED_OFF);
        } else {
            led_set(LED_VALVE_MANUAL, LED_ON);
        }

        // VALVE control state machine
        if (valve_lk_command == VALVE_COMMAND_CLOSE) {
            if (wmc_state != WMC_STATE_LEAKAGE_CLOSE) {
                wmc_state = WMC_STATE_LEAKAGE_CLOSE;
#ifdef DEBUG_VALVE
                printf("leakage detection, close\r\n");
#endif // DEBUG_VALVE
                valve_close();
            }
        } else if (valve_pwr_command == VALVE_COMMAND_CLOSE) {
            if (wmc_state != WMC_STATE_BATLOW_CLOSE) {
                wmc_state = WMC_STATE_BATLOW_CLOSE;
#ifdef DEBUG_VALVE
                printf("battery low close\r\n");
#endif // DEBUG_VALVE
                valve_close();
            }
        } else if (valve_manual_command == VALVE_COMMAND_CLOSE) {
            if (wmc_state != WMC_STATE_MANUAL_CLOSE) {
                wmc_state = WMC_STATE_MANUAL_CLOSE;
#ifdef DEBUG_VALVE
                printf("manual close\r\n");
#endif // DEBUG_VALVE
                valve_close();
            }
        } else if (valve_manual_command == VALVE_COMMAND_OPEN) {
            if (wmc_state != WMC_STATE_MANUAL_OPEN) {
                wmc_state = WMC_STATE_MANUAL_OPEN;
#ifdef DEBUG_VALVE
                printf("manual open\r\n");
#endif // DEBUG_VALVE
                valve_open();
            }
        } else if (valve_remote_command == VALVE_COMMAND_CLOSE) {
            if (wmc_state != WMC_STATE_REMOTE_CLOSE) {
                wmc_state = WMC_STATE_REMOTE_CLOSE;
#ifdef DEBUG_VALVE
                printf("remote close\r\n");
#endif // DEBUG_VALVE
                valve_close();
            }
        } else if (valve_remote_command == VALVE_COMMAND_OPEN) {
            if (wmc_state != WMC_STATE_REMOTE_OPEN) {
                wmc_state = WMC_STATE_REMOTE_OPEN;
#ifdef DEBUG_VALVE
                printf("remote open\r\n");
#endif // DEBUG_VALVE
                valve_open();
            }
        } else if (valve_remote_command == VALVE_COMMAND_UNDEFINED) {
            if (wmc_state != WMC_STATE_WTG_CLOSE) {
                wmc_state = WMC_STATE_WTG_CLOSE;
#ifdef DEBUG_VALVE
                printf("alive watchdog close\r\n");
#endif // DEBUG_VALVE
                valve_close();
            }
        }

        // schedule VALVE task
        if (TASK_SCHED_CHECK(task_valve)) {
            uint8_t valve_state;

            TASK_SCHED_SYNC(task_valve);

            valve_poll();

            valve_state = valve_get_state();

            if (wmc_state == WMC_STATE_BATLOW_CLOSE) {
                // switch off VALVE OPEN/CLOSE LEDs when battery low
                led_set(LED_VALVE_CLOSE, LED_OFF);
                led_set(LED_VALVE_OPEN, LED_OFF);
            } else if (valve_state != last_valve_state) {
                // update VALVE status LEDs
                switch (valve_state) {
                case VALVE_STATE_WAIT_CLOSE:
                    led_set(LED_VALVE_CLOSE, LED_FLASH);
                    led_set(LED_VALVE_OPEN, LED_OFF);
                    break;
                case VALVE_STATE_CLOSE:
                    led_set(LED_VALVE_CLOSE, LED_ON);
                    led_set(LED_VALVE_OPEN, LED_OFF);
                    buz_valve_close();
                    break;
                case VALVE_STATE_WAIT_OPEN:
                    led_set(LED_VALVE_CLOSE, LED_OFF);
                    led_set(LED_VALVE_OPEN, LED_FLASH);
                    break;
                case VALVE_STATE_OPEN:
                    led_set(LED_VALVE_CLOSE, LED_OFF);
                    led_set(LED_VALVE_OPEN, LED_ON);
                    buz_valve_open();
                    break;
                case VALVE_STATE_ERROR:
                default:
                    led_set(LED_VALVE_CLOSE, LED_FLASH);
                    led_set(LED_VALVE_OPEN, LED_FLASH_INV);
                    buz_valve_error();
                }
                last_valve_state = valve_state;
            }
        }

        // schedule LEDs task
        if (TASK_SCHED_CHECK(task_led)) {
            TASK_SCHED_SYNC(task_led);
            led_poll();
        }

        // low-priority task
        if(TASK_SCHED_CHECK(task_low)) {
            TASK_SCHED_SYNC(task_low);

            // leakage sensor status indication
            if (wmc_state == WMC_STATE_LEAKAGE_CLOSE) {
                if (lk_get_state() == LK_STATUS_LEAK) {
                    led_set(LED_LEAKAGE, LED_BLINK);
                    buz_leak();
                } else {
                    led_set(LED_LEAKAGE, LED_ON);
                }
            } else {
                led_set(LED_LEAKAGE, LED_OFF);
            }

            // indication of the end of washing
            if (wmc_state == WMC_STATE_REMOTE_OPEN) {
                if (washing_time < WASHING_TIME_THR) washing_time++;
            } else if (wmc_state == WMC_STATE_REMOTE_CLOSE) {
                if (washing_time >= WASHING_TIME_THR) {
                    if (valve_get_state() == VALVE_STATE_CLOSE) {
                        if (washing_time % 2) {
                            buz_finish();
                        }
                        if (washing_time++ >= WASHING_TIME_THR + (WASHING_TIME_ALARM * 2)) {
                            washing_time = 0;
                        }
                    }
                } else {
                    washing_time = 0;
                }
            } else {
                washing_time = 0;
            }
        }

        // reset WDT
        wdt_reset();
    }

    return 0;
}
