#include <avr/io.h>

#include "power.h"

#ifdef DEBUG
#include <stdio.h>
#endif // DEBUG


#if defined (__AVR_ATmega16__)
#   undef ADC_vect
#   define ADC_vect _VECTOR(28)
#endif // defined (__AVR_ATmega16__)


#define PWR_ADC_LINE            7
#define PWR_ADC_BAT             6
#define PWR_CHARGE_PIN          1

#define PWR_AVG_LINE            32
#define PWR_AVG_BAT             32

#define PWR_LINE_THRESHOLD      700
#define PWR_BATALARM_THRESHOLD  580
#define PWR_BATLOW_THRESHOLD    530



static int16_t pwr_line_value;
static int16_t pwr_bat_value;
static int8_t pwr_state;



uint8_t pwr_get_charge();

static int16_t pwr_adc(uint8_t mux) {
    ADMUX = _BV(REFS0) | mux;       // REF from AVCC and mux channel
    ADCSRA |= _BV(ADSC);            // start conversion
    while (ADCSRA & _BV(ADSC));     // wait for conversion to complete
    return ADCW;                    // return ADC value
}


static void pwr_fsm() {
    switch (pwr_state) {
    case PWR_STATE_ONLINE:
        if (pwr_line_value < PWR_LINE_THRESHOLD)
            pwr_state = PWR_STATE_BAT;
        else if (pwr_get_charge())
            pwr_state = PWR_STATE_ONLINE_CHARGE;
        break;
    case PWR_STATE_ONLINE_CHARGE:
        if (pwr_line_value < PWR_LINE_THRESHOLD)
            pwr_state = PWR_STATE_BAT;
        else if (!pwr_get_charge())
            pwr_state = PWR_STATE_ONLINE;
        break;
    case PWR_STATE_BAT:
        if (pwr_line_value >= PWR_LINE_THRESHOLD)
            pwr_state = PWR_STATE_ONLINE;
        else if (pwr_bat_value < PWR_BATALARM_THRESHOLD)
            pwr_state = PWR_STATE_BATALARM;
        break;
    case PWR_STATE_BATALARM:
        if (pwr_line_value >= PWR_LINE_THRESHOLD)
            pwr_state = PWR_STATE_ONLINE;
        else if (pwr_bat_value < PWR_BATLOW_THRESHOLD)
            pwr_state = PWR_STATE_BATLOW;
        break;
    case PWR_STATE_BATLOW:
        if (pwr_line_value >= PWR_LINE_THRESHOLD)
            pwr_state = PWR_STATE_ONLINE;
        break;
    default:
        pwr_state = PWR_STATE_ONLINE;
    }
}


//
// interface functions
//

void pwr_init() {
    // set charge status pin to IN
    DDRC &= ~_BV(PWR_CHARGE_PIN);

    // enable ADC and set division factor to 128
    ADCSRA = _BV(ADEN) | _BV(ADPS2) | _BV(ADPS1) | _BV(ADPS0);

    // get initial values of LINE and BAT channels
    pwr_line_value = pwr_adc(PWR_ADC_LINE);
    pwr_bat_value = pwr_adc(PWR_ADC_BAT);

    pwr_state = PWR_STATE_ONLINE;
}


uint8_t pwr_get_state() {
    return pwr_state;
}


uint8_t pwr_get_charge() {
    return (PINC & _BV(PWR_CHARGE_PIN)) ? 0 : 1;
}


int16_t pwr_get_line() {
    return pwr_line_value;
}


int16_t pwr_get_bat() {
    return pwr_bat_value;
}


void pwr_poll() {
    int16_t diff;

    // average LINE values
    diff = pwr_adc(PWR_ADC_LINE) - pwr_line_value;
    diff += diff > 0 ? (PWR_AVG_LINE - 1) : -(PWR_AVG_LINE - 1);
    pwr_line_value += diff / PWR_AVG_LINE;

    // average BATTERY values
    diff = pwr_adc(PWR_ADC_BAT) - pwr_bat_value;
    diff += diff > 0 ? (PWR_AVG_BAT - 1) : -(PWR_AVG_BAT - 1);
    pwr_bat_value += diff / PWR_AVG_BAT;

    pwr_fsm();
}
