#include <avr/io.h>

#include "valve.h"

#ifdef DEBUG
#include <stdio.h>
#endif // DEBUG



#define VALVE_PIN_MOTOR_A       _BV(7)
#define VALVE_PIN_MOTOR_B       _BV(6)
#define VALVE_PIN_SENSOR_CLOSE  _BV(4)
#define VALVE_PIN_SENSOR_OPEN   _BV(5)

#define VALVE_SENSOR_ERROR      0
#define VALVE_SENSOR_UNDEFINED  1
#define VALVE_SENSOR_CLOSE      2
#define VALVE_SENSOR_OPEN       3

#define VALVE_MOTOR_TIMEOUT     80      // 8s * 1000 / 100ucs



/* VALVE state */
static uint8_t valve_command;
static uint8_t valve_state;
static uint16_t valve_wait;



/* motor control functions */
static void motor_forward() {
    PORTD &= ~VALVE_PIN_MOTOR_A;
    PORTD |= VALVE_PIN_MOTOR_B;
}

static void motor_reverse() {
    PORTD &= ~VALVE_PIN_MOTOR_B;
    PORTD |= VALVE_PIN_MOTOR_A;
}

static void motor_stop() {
    PORTD &= ~(VALVE_PIN_MOTOR_A | VALVE_PIN_MOTOR_B);
}


/* sensor control functions */
static int sensor_get() {
    uint8_t state = PIND;
    uint8_t state_close = (state & VALVE_PIN_SENSOR_CLOSE) == 0;
    uint8_t state_open = (state & VALVE_PIN_SENSOR_OPEN) == 0;

    if (state_close && state_open) {
        return VALVE_SENSOR_ERROR;
    } else if (state_close && !state_open) {
        return VALVE_SENSOR_CLOSE;
    } else if (!state_close && state_open) {
        return VALVE_SENSOR_OPEN;
    } else {
        return VALVE_SENSOR_UNDEFINED;
    }
}



/*
 * interface functions
 */

void valve_init() {
    DDRD |= VALVE_PIN_MOTOR_A | VALVE_PIN_MOTOR_B;
    DDRD &= ~(VALVE_PIN_SENSOR_CLOSE | VALVE_PIN_SENSOR_OPEN);

    // stop motor
    motor_stop();

    valve_state = VALVE_STATE_ERROR;
    valve_command = VALVE_COMMAND_UNDEFINED;
}


void valve_close() {
    if (valve_command != VALVE_COMMAND_CLOSE || valve_state == VALVE_STATE_ERROR) {
        valve_command = VALVE_COMMAND_CLOSE;
        valve_wait = 0;
        valve_state = VALVE_STATE_WAIT_CLOSE;
    }
}


void valve_open() {
    if (valve_command != VALVE_COMMAND_OPEN || valve_state == VALVE_STATE_ERROR) {
        valve_command = VALVE_COMMAND_OPEN;
        valve_wait = 0;
        valve_state = VALVE_STATE_WAIT_OPEN;
    }
}


uint8_t valve_get_state() {
    return valve_state;
}


void valve_poll() {
    uint8_t sensor_state;

    if (valve_state == VALVE_STATE_ERROR) {
        return;
    }

    sensor_state = sensor_get();

    if (valve_command == VALVE_COMMAND_CLOSE) {
        if (sensor_state != VALVE_SENSOR_CLOSE) {
            if (valve_wait++ >= VALVE_MOTOR_TIMEOUT) {
                motor_stop();
                valve_state = VALVE_STATE_ERROR;
#ifdef DEBUG
                printf("valve: close error\r\n");
#endif // DEBUG
            } else {
                motor_forward();
            }
        } else {
            motor_stop();
#ifdef DEBUG
            if (valve_state != VALVE_STATE_CLOSE) printf("valve: close %d\r\n", valve_wait);
#endif // DEBUG
            valve_state = VALVE_STATE_CLOSE;
        }
    } else if (valve_command == VALVE_COMMAND_OPEN) {
        if (sensor_state != VALVE_SENSOR_OPEN) {
            if (valve_wait++ >= VALVE_MOTOR_TIMEOUT) {
                motor_stop();
                valve_state = VALVE_STATE_ERROR;
#ifdef DEBUG
                printf("valve: open error\r\n");
#endif // DEBUG
            } else {
                motor_reverse();
            }
        } else {
            motor_stop();
#ifdef DEBUG
            if (valve_state != VALVE_STATE_OPEN) printf("valve: open %d\r\n", valve_wait);
#endif // DEBUG
            valve_state = VALVE_STATE_OPEN;
        }
    } else {
        motor_stop();
#ifdef DEBUG
        if (valve_state != VALVE_STATE_ERROR) printf("valve: unknown commnad\r\n");
#endif // DEBUG
        valve_state = VALVE_STATE_ERROR;
    }
}
