/*
 * Washer Machine Protector (WMP)
 * Monitor module firmware
 *
 * (C)2017 St(u)dio of Computer Games
 * http://stdio.ru/
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>
#include <string.h>

#include "VirtualWire.h"

#include "wmp.h"
#include "i2c_candy.h"



#define VW_BAUD                 600
#define I2CC_MSG_NUM            3
#define STATUS_THRESHOLD        3


#define WMP_STATUS_UNDEF        0
#define WMP_STATUS_POWER_OFF    1
#define WMP_STATUS_STOP         2
#define WMP_STATUS_START        3

#define MSG_FLAG_NONE           0
#define MSG1_FLAG_POWER_ON      _BV(0)
#define MSG2_FLAG_POWER_ON      _BV(1)
#define MSG3_FLAG_START         _BV(2)



//
// debug section
//

#ifdef DEBUG

#include <stdio.h>

static void uart_init() {
    UCSRB |= _BV(RXEN) | _BV(TXEN);
	UCSRC |= _BV(URSEL) | _BV(UCSZ0) | _BV(UCSZ1);

//#define BAUD 230400
#define BAUD 57600
#include <util/setbaud.h>
   UBRRH = UBRRH_VALUE;
   UBRRL = UBRRL_VALUE;
#if USE_2X
   UCSRA |= _BV(U2X);
#else
   UCSRA &= ~_BV(U2X);
#endif
}

static int uart_puthcar(char c, FILE *stream) {
	while ((UCSRA & _BV(UDRE)) == 0) {};
	UDR = c;
	return 0;
}

static FILE uart_stream = FDEV_SETUP_STREAM(uart_puthcar, NULL, _FDEV_SETUP_RW);


void uart_dump_buffer(uint8_t *data, size_t len) {
    size_t i;

    printf("%d:", len);
    for (i = 0; i < len; i++) {
        printf(" %02x", data[i]);
    }
    printf("\r\n");
}


static uint8_t i2cc_csum[4];

static void msg_dump(i2cc_msg_t *i2cc_msg) {
    uint8_t i;
    printf("%d: ", i2cc_msg->msg.idx);
    for (i = 0; i < 8; i++) {
        printf("%02x ", i2cc_msg->msg.data[i]);
    }
    printf("\r\n");
}

#endif // DEBUG



//
// main section
//

wmp_msg_t wmp_msg = { WMP_ADDR_CONTROLLER, WMP_ADDR_MONITOR, 0 };



static void wm_send_status(uint8_t monitor_status) {
    if (monitor_status == WMP_STATUS_POWER_OFF || monitor_status == WMP_STATUS_UNDEF) {
        return;
    }

    wmp_msg.status = WMP_MSG_STATUS_ALIVE;
    if (monitor_status == WMP_STATUS_START) {
        wmp_msg.status |= WMP_MSG_STATUS_VALVE;
    }

#ifdef DEBUG
    printf("wm_send_status(): ");
    uart_dump_buffer((uint8_t *)&wmp_msg, sizeof(wmp_msg));
#endif // DEBUG

    vw_send((uint8_t *)&wmp_msg, sizeof(wmp_msg));
    vw_wait_tx(); // Wait until the whole message is gone
}



// ALIVE timer
volatile uint8_t alive_cnt;
volatile uint8_t alive_sync;


ISR(TIMER2_COMP_vect) {
    if(++alive_cnt >= 244) {
        alive_cnt = 0;
        alive_sync = 1;
    }
}


static uint8_t wmp_get_flags(i2cc_msg_t *msg) {
    if (msg->msg.data[0] == 0x0 && msg->msg.data[1] == 0x0 &&
            msg->msg.data[2] == 0x0 && msg->msg.data[3] == 0x0 &&
            msg->msg.data[4] == 0x0 && msg->msg.data[5] == 0x0 &&
            msg->msg.data[6] == 0x0 && msg->msg.data[7] == 0x0 ) {
        if (msg->msg.idx == 1 || msg->msg.idx == 2) {
            return MSG_FLAG_NONE;
        }
    } else {
        if (msg->msg.idx == 1) {
            return MSG1_FLAG_POWER_ON;
        } else if (msg->msg.idx == 2) {
            return MSG2_FLAG_POWER_ON;
        }
    }

    if (msg->msg.idx == 3) {
        if (msg->msg.data[4] == 0x1 && msg->msg.data[5] == 0x1 &&
                msg->msg.data[6] == 0x1 && msg->msg.data[7] == 0x1) {
            return MSG3_FLAG_START;
        } else if (msg->msg.data[4] == 0x0 && msg->msg.data[5] == 0x0 &&
                msg->msg.data[6] == 0x0 && msg->msg.data[7] == 0x0) {
            return MSG_FLAG_NONE;
        }
    }

    return MSG_FLAG_NONE;
}



//
// Entry point
//

int main(void) {
#ifdef DEBUG
    uint8_t mcucsr = MCUCSR;
    MCUCSR = 0;
#endif // DEBUG

    // start WDT
    wdt_enable(WDTO_1S);
    wdt_reset();

#ifdef DEBUG
    uart_init();
    stdout = &uart_stream;
    printf("WMP Monitor start, MCUCSR 0x%x\r\n", mcucsr);
#endif // DEBUG
    cli();

    vw_setup(VW_BAUD);
    i2cc_init();

    // ALIVE timer (TIMER2)
    TCCR2 = _BV(WGM21) | _BV(CS22) | _BV(CS21); // set CTC mode, prescaler 1/256
    OCR2  = 0xff;
    TIMSK |= _BV(OCIE2);                        //  TIMER2, Compare Interrupt Enable

    alive_sync = 0;
    alive_cnt = 0;

    uint8_t msg_buf_flags = 0;
    uint8_t msg_flags = 0;

    uint8_t monitor_status = WMP_STATUS_UNDEF;
    uint8_t wm_last_status = WMP_STATUS_UNDEF;
    uint8_t wm_last_status_cnt = 0;

    // enable interrupts
    sei();

    i2cc_start();

#ifdef DEBUG
    i2cc_csum[0] = i2cc_csum[2] = i2cc_csum[2] = i2cc_csum[3] = 0;
#endif // DEDUG

    while(1) {
        i2cc_msg_t *i2cc_msg;

        if ((i2cc_msg = i2cc_get_msg()) != NULL) {
            if (i2cc_msg->msg.idx < 1 || i2cc_msg->msg.idx > I2CC_MSG_NUM) {
                continue;
            }

            // process message
            msg_flags |= wmp_get_flags(i2cc_msg);
            msg_buf_flags |= _BV(i2cc_msg->msg.idx - 1);

            // get all 3 messages
            if ((msg_buf_flags & (_BV(I2CC_MSG_NUM) - 1)) == (_BV(I2CC_MSG_NUM) - 1)) {
                uint8_t wm_status, monitot_new_status;

                // translate flags to washing machine status
                if (msg_flags & MSG3_FLAG_START) {
                        wm_status = WMP_STATUS_START;
                } else if ((msg_flags & MSG1_FLAG_POWER_ON) || (msg_flags & MSG2_FLAG_POWER_ON)) {
                        wm_status = WMP_STATUS_STOP;
                } else {
                        wm_status = WMP_STATUS_POWER_OFF;
                }

                // update last washing machine status
                if (wm_last_status != wm_status) {
                    wm_last_status = wm_status;
#ifdef DEBUG
                    printf("wm: flags %x, status %d, last_status %d, last_status_cnt %d\r\n",
                            msg_flags, wm_status, wm_last_status, wm_last_status_cnt);
#endif // DEBUG
                    wm_last_status_cnt = 0;
                }

                // monitor state switch decision
                monitot_new_status = WMP_STATUS_UNDEF;
                switch (monitor_status) {
                case WMP_STATUS_POWER_OFF:
                    if (wm_status == WMP_STATUS_STOP) {
                        monitot_new_status = WMP_STATUS_STOP;
                    }
                    break;
                case WMP_STATUS_STOP:
                    if (wm_status == WMP_STATUS_POWER_OFF)  {
                        monitot_new_status = WMP_STATUS_POWER_OFF;
                    } else if (wm_status == WMP_STATUS_START) {
                        monitot_new_status = WMP_STATUS_START;
                    }
                    break;
                case WMP_STATUS_START:
                    if (wm_status == WMP_STATUS_POWER_OFF)  {
                        monitot_new_status = WMP_STATUS_POWER_OFF;
                    } else if (wm_status == WMP_STATUS_STOP &&
                                wm_last_status_cnt > STATUS_THRESHOLD) {
                        monitot_new_status = WMP_STATUS_STOP;
                    }
                    break;
                case WMP_STATUS_UNDEF:
                default:
                    monitot_new_status = wm_status;
                }

                // update monitor status
                if (monitot_new_status != WMP_STATUS_UNDEF && monitor_status != monitot_new_status) {
#ifdef DEBUG
                    printf("mon: status %d, new_status %d, wm_status %d, wm_last_status_cnt %d\r\n",
                        monitor_status, monitot_new_status, wm_status, wm_last_status_cnt);
#endif // DEBUG

                    monitor_status = monitot_new_status;

                    // disable ALIVE timer and INT0 interrupt
                    TIMSK &= ~_BV(OCIE2);

                    wm_send_status(monitor_status);

                    // enable and restart ALIVE times
                    alive_cnt = 0;
                    alive_sync = 0;
                    TCNT2 = 0;
                    TIMSK |= _BV(OCIE2);
                }

                msg_flags = 0;
                msg_buf_flags = 0;
            }

#ifdef DEBUG
            if (i2cc_msg->msg.idx >= 1 && i2cc_msg->msg.idx <= I2CC_MSG_NUM) {
                if (i2cc_msg->msg.csum != i2cc_csum[i2cc_msg->msg.idx - 1]) {
                    i2cc_csum[i2cc_msg->msg.idx - 1] = i2cc_msg->msg.csum;
                    printf("msg: msg_flags 0x%x, msg_buf_flags 0x%x ", msg_flags, msg_buf_flags);
                    msg_dump(i2cc_msg);
                }
            }
#endif // DEBUG

            // restart I2C listener
            i2cc_start();

            // reset WDT
            wdt_reset();
        }

        if (alive_sync) {
            if (wm_last_status_cnt < 255) wm_last_status_cnt++;
            wm_send_status(monitor_status);
            alive_sync = 0;
        }
    }

    return 0;
}
